/***************************************************************************
 *   Copyright (C) 2007 by Lisandro Damián Nicanor Pérez Meyer             *
 *   perezmeyer@gmail.com                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <QPainter>
#include <QPaintEvent>
#include <QDebug>
#include <math.h>
#include "scene.h"
#include "cube.h"
#include "matrix.h"

#define PI 3.141592654

Scene::Scene(QWidget * parent) : QWidget(parent)
{
	selectedFigure = -1;

	perspective = false;

	image = new QImage(width(), height(), QImage::Format_RGB32);

	xDisplacement = 0.0;
	yDisplacement = 0.0;
	zDisplacement = 0.0;
	xAngle = 0.0;
	yAngle = 0.0;
	zAngle = 0.0;
	xScale = 1.0;
	yScale = 1.0;
	zScale = 1.0;
}

Scene::~Scene()
{
	// Empty the figures list
	qDeleteAll(figures.begin(), figures.end());
	figures.clear();
	// Delete the image
	delete image;
}

/**
	We will first delete the previous QImage and create a new one with the
	current Scene's width and height. In this way we do not have to worry about
	resizes nor to clean the image in each frame.\n
	Allmost the same thing happens with the Z-Buffer matrix. But in this case is
	much easier to deploy in this way.\n
	\n
	Note: perhaps cleaning the image is faster that deleting and creating a new
	one, and only take care of resize events when it's appropiate. I have not
	tested this.
*/
void Scene::draw()
{
	// Delete the previous image
	delete image;
	// Create a new one
	image = new QImage(width(), height(), QImage::Format_RGB32);

	// Create the Z-Buffer
	Matrix * zbuffer = new Matrix(width(), height());

	// Create and apply the different transformation matrices
	Matrix4X4 projection = getProjectionMatrix();
	Matrix4X4 scale = getScaleMatrix();
	Matrix4X4 rotation = getRotationMatrix();
	Matrix4X4 translation = getTranslationMatrix();

	/// FIXME
	/**
	  This code has been commented out by a comment by my teacher. I do not
	  really understand why it works.
	*/
// 	if(perspective)
// 	{
// 		Matrix4X4 temp1 = projection.multiplyToTheRight(scale);
// 		scale = temp1;
// 	}

	Matrix4X4 temp2 = rotation.multiplyToTheRight(scale);
	Matrix4X4 matrix = translation.multiplyToTheRight(temp2);

	// Apply the matrices to the figures
	for(int i=0; i< figures.size(); i++)
	{
		figures[i]->draw(matrix, zbuffer, perspective, image);
	}

	// Delete the Z-Buffer
	delete zbuffer;
}

void Scene::addCube()
{
	Cube * cube = new Cube();
	figures.append(cube);
	update();
}

void Scene::perspectiveProyection(int state)
{
	if(state == Qt::Checked)
		perspective = true;
	else
		perspective = false;
	update();
}

void Scene::figureSelected(int figure)
{
	if(figure != -1)
	{
		emit changeParameters(figures[figure]->getXDisplacement(),
		                      figures[figure]->getYDisplacement(),
		                      figures[figure]->getAngleOfRotationX(),
		                      figures[figure]->getAngleOfRotationY(),
		                      figures[figure]->getAngleOfRotationZ(),
		                      figures[figure]->getXScale(),
		                      figures[figure]->getYScale());
	}
	selectedFigure = figure;
}

void Scene::changeXDisplacement(int displacement)
{
	if(selectedFigure != -1)
	{
		figures[selectedFigure]->setXDisplacement((double) displacement / 500.0);
		update();
	}
}

void Scene::changeYDisplacement(int displacement)
{
	if(selectedFigure != -1)
	{
		figures[selectedFigure]->setYDisplacement((double) displacement / 500.0);
		update();
	}
}

void Scene::changeZDisplacement(int displacement)
{
	if(selectedFigure != -1)
	{
		figures[selectedFigure]->setZDisplacement((double) displacement / 500.0);
		update();
	}
}

void Scene::changeAngleOfRotationX(int angle)
{
	if(selectedFigure != -1)
	{
		// Maths functions are in radians, but sliders are in degrees.
		double radians = (int)angle * PI / 180.0;
		figures[selectedFigure]->setAngleOfRotationX(radians);
		update();
	}
}

void Scene::changeAngleOfRotationY(int angle)
{
	if(selectedFigure != -1)
	{
		// Maths functions are in radians, but sliders are in degrees.
		double radians = (int)angle * PI / 180.0;
		figures[selectedFigure]->setAngleOfRotationY(radians);
		update();
	}
}

void Scene::changeAngleOfRotationZ(int angle)
{
	if(selectedFigure != -1)
	{
		// Maths functions are in radians, but sliders are in degrees.
		double radians = (int)angle * PI / 180.0;
		figures[selectedFigure]->setAngleOfRotationZ(radians);
		update();
	}
}

void Scene::changeXScale(int scale)
{
	if(selectedFigure != -1)
	{
		figures[selectedFigure]->setXScale((double) scale / 50.0);
		update();
	}
}

void Scene::changeYScale(int scale)
{
	if(selectedFigure != -1)
	{
		figures[selectedFigure]->setYScale((double) scale / 50.0);
		update();
	}
}

void Scene::changeZScale(int scale)
{
	if(selectedFigure != -1)
	{
		figures[selectedFigure]->setZScale((double) scale / 50.0);
		update();
	}
}

void Scene::changeSceneXDisplacement(int displacement)
{
	xDisplacement = (double) displacement / 500.0;
	update();
}

void Scene::changeSceneYDisplacement(int displacement)
{
	yDisplacement = (double) displacement / 500.0;
	update();
}

void Scene::changeSceneZDisplacement(int displacement)
{
	zDisplacement = (double) displacement / 500.0;
	update();
}

void Scene::changeSceneAngleOfRotationX(int angle)
{
	xAngle = (int)angle * PI / 180.0;
	update();
}

void Scene::changeSceneAngleOfRotationY(int angle)
{
	yAngle = (int)angle * PI / 180.0;
	update();
}

void Scene::changeSceneAngleOfRotationZ(int angle)
{
	zAngle = (int)angle * PI / 180.0;
	update();
}

void Scene::changeSceneXScale(int scale)
{
	xScale = (double) scale / 50.0;
	update();
}

void Scene::changeSceneYScale(int scale)
{
	yScale = (double) scale / 50.0;
	update();
}

void Scene::changeSceneZScale(int scale)
{
	zScale = (double) scale / 50.0;
	update();
}

void Scene::paintEvent(QPaintEvent *)
{
	QPainter painter(this);

	// Draw the objects in the image
	draw();

	// Draw the image
	painter.drawImage(0, 0, *image);
}

Matrix4X4 Scene::getProjectionMatrix()
{
	Matrix4X4 matrix;

	matrix.setValue(0, 0, 1.0);
	matrix.setValue(0, 1, 0.0);
	matrix.setValue(0, 2, 0.0);
	matrix.setValue(0, 3, 0.0);

	matrix.setValue(1, 0, 0.0);
	matrix.setValue(1, 1, 1.0);
	matrix.setValue(1, 2, 0.0);
	matrix.setValue(1, 3, 0.0);

	matrix.setValue(2, 0, 0.0);
	matrix.setValue(2, 1, 0.0);
	matrix.setValue(2, 2, -11.0/9.0);
	matrix.setValue(2, 3, -20.0/9.0);

	matrix.setValue(3, 0, 0.0);
	matrix.setValue(3, 1, 0.0);
	matrix.setValue(3, 2, -1.0);
	matrix.setValue(3, 3, 0.0);

	return matrix;
}

Matrix4X4 Scene::getScaleMatrix()
{
	Matrix4X4 matrix;

	matrix.setValue(0, 0, xScale);
	matrix.setValue(0, 1, 0.0);
	matrix.setValue(0, 2, 0.0);
	matrix.setValue(0, 3, 0.0);

	matrix.setValue(1, 0, 0.0);
	matrix.setValue(1, 1, yScale);
	matrix.setValue(1, 2, 0.0);
	matrix.setValue(1, 3, 0.0);

	matrix.setValue(2, 0, 0.0);
	matrix.setValue(2, 1, 0.0);
	matrix.setValue(2, 2, zScale);
	matrix.setValue(2, 3, 0.0);

	matrix.setValue(3, 0, 0.0);
	matrix.setValue(3, 1, 0.0);
	matrix.setValue(3, 2, 0.0);
	matrix.setValue(3, 3, 1.0);

	return matrix;
}

Matrix4X4 Scene::getRotationMatrix()
{
	Matrix4X4 matrix;

	matrix.setValue(0, 0, cos(yAngle) * cos(zAngle));
	matrix.setValue(0, 1, cos(yAngle) * -sin(zAngle));
	matrix.setValue(0, 2, -sin(yAngle));
	matrix.setValue(0, 3, 0.0);

	matrix.setValue(1, 0, -sin(xAngle) * sin(yAngle) *
	                cos(zAngle) + cos(xAngle) *
	                sin(zAngle));
	matrix.setValue(1, 1, sin(xAngle) * sin(yAngle) *
	                sin(zAngle) + cos(xAngle) *
	                cos(zAngle));
	matrix.setValue(1, 2, -sin(xAngle) * cos(yAngle));
	matrix.setValue(1, 3, 0.0);

	matrix.setValue(2, 0, cos(xAngle) * sin(yAngle) *
	                cos(zAngle) + sin(xAngle) *
	                sin(zAngle));
	matrix.setValue(2, 1, cos(xAngle) * sin(yAngle) *
	                -sin(zAngle) + sin(xAngle) *
	                cos(zAngle));
	matrix.setValue(2, 2, cos(xAngle) * cos(yAngle));
	matrix.setValue(2, 3, 0.0);

	matrix.setValue(3, 0, 0.0);
	matrix.setValue(3, 1, 0.0);
	matrix.setValue(3, 2, 0.0);
	matrix.setValue(3, 3, 1.0);

	return matrix;
}

Matrix4X4 Scene::getTranslationMatrix()
{
	Matrix4X4 matrix;

	matrix.setValue(0, 0, 1.0);
	matrix.setValue(0, 1, 0.0);
	matrix.setValue(0, 2, 0.0);
	matrix.setValue(0, 3, xDisplacement);

	matrix.setValue(1, 0, 0.0);
	matrix.setValue(1, 1, 1.0);
	matrix.setValue(1, 2, 0.0);
	matrix.setValue(1, 3, yDisplacement);

	matrix.setValue(2, 0, 0.0);
	matrix.setValue(2, 1, 0.0);
	matrix.setValue(2, 2, 1.0);
	matrix.setValue(2, 3, zDisplacement);

	matrix.setValue(3, 0, 0.0);
	matrix.setValue(3, 1, 0.0);
	matrix.setValue(3, 2, 0.0);
	matrix.setValue(3, 3, 1.0);

	return matrix;
}



